/* fitshdr - utility to read/modify FITS-headers
 *
 * $Id: fitshdr.c,v 1.69 2006/06/02 04:08:39 mik Exp $
 *
 */

#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <getopt.h>

#include "fits.h"
#include "version.h"

#define OPT_TEST		(0x1)
#define OPT_LINENO		(0x2)
#define OPT_HDROFF		(0x4)
#define OPT_EMPTYLINE		(0x8)
#define OPT_BLKOFF		(0x10)
#define OPT_DATAOFF		(0x20)
#define OPT_IGNOLDHDR		(0x40)
#define OPT_REFMTCARD		(0x80)
#define OPT_MODIFY_PRINCIPAL	(0x100)
#define OPT_UNDEF		(0x8000)
#define OPT_DEFAULT	(OPT_REFMTCARD|OPT_BLKOFF|OPT_HDROFF|OPT_DATAOFF)

#define PNAME		"fitshdr"
#define COMMENT_PROMPT	";"
#define ADDPROMPT	COMMENT_PROMPT";; "


#define CHKLVL_DEFAULT		(0)
/* dump */
#define CHKLVL_E_TRUNC_HDR	(0)
#define CHKLVL_E_MAGIC_NUM	(0)
#define CHKLVL_W_CARD_FORMAT	(1)
#define CHKLVL_E_CARD_FORMAT	(CHKLVL_W_CARD_FORMAT+1)
#define CHKLVL_W_NONASCII	(CHKLVL_W_CARD_FORMAT)
#define CHKLVL_E_NONASCII	(CHKLVL_E_CARD_FORMAT+1)
/* edit */
#define CHKLVL_E_UNSORTED	(0)
#define CHKLVL_E_NEWCARD_FORMAT	(0)
#define CHKLVL_E_TRUNC_COMMENT	(1)
#define CHKLVL_E_OLDCARD_FORMAT	(2)
#define CHKLVL_CHK_NEWCARD	(-2)

#define VERBOSE_LVL_ERROR	(0)
#define VERBOSE_LVL_WARNING	(VERBOSE_LVL_ERROR+1)
#define VERBOSE_LVL_DEBUG	(VERBOSE_LVL_WARNING+1)
#define VERBOSE_LVL_SILENT	(-1)
#define VERBOSE_LVL_DEFAULT	(VERBOSE_LVL_ERROR)

int ed_header(const char *ifn, const char *hfn, const int opts, const int pedantic, const int verbose);
int dump_header(const char *ifn, const int opts, const int pedantic, const int verbose);

int
main(int ac, char **av){
  char *ifn=NULL, *hfn=NULL;
  int c, rc=0, verbose=INT_MIN, pedantic=INT_MIN;
  int opts=OPT_UNDEF, is_edit_mode=-1;
  static struct option long_options[] = {
    {"help", 0, 0, 'h'},
    {"verbose", 0, 0, 'v'},
    {"silent", 0, 0, 'q'},
    {"test", 0, 0, 'n'},
    {"empty-lines", 0, 0, 1025},
    {"no-empty-lines", 0, 0, 1029},
    {"line-offsets", 0, 0, 1026},
    {"no-line-offsets", 0, 0, 1030},
    {"header-offsets", 0, 0, 1027},
    {"no-header-offsets", 0, 0, 1031},
    {"block-offsets", 0, 0, 1028},
    {"no-block-offsets", 0, 0, 1032},
    {"no-data-offsets", 0, 0, 1034},
    {"data-offsets", 0, 0, 1033},
    {"reformat-cards", 0, 0, 1035},
    {"no-reformat-cards", 0, 0, 1036},
    {"discard-old-headers", 0, 0, 1037},
    {"edit", 0, 0, 1038},
    {"modify-principal-keywords", 0, 0, 1039},
    {"pedantic", 0, 0, 'p'},
    {"force", 0, 0, 'f'},
    {"file", 1, 0, 'F'},
    {"fits-file", 1, 0, 'F'},
    {"header", 1, 0, 'H'},
    {"version", 0, 0, 'V'},
    {0, 0, 0, 0}
  };
  int option_index = 0;

  for(;;){
    c = getopt_long(ac, av, "fpnvVqhF:H:", long_options, &option_index);
    if(c == -1) break;
    switch(c){
      case 'V':
        fprintf(stderr, "%s: version %s\n", PNAME, FITSHDR_VERSION);
        return(0);
      case 1039:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_MODIFY_PRINCIPAL;
        break;
      case 1038:
        is_edit_mode=1;
        break;
      case 1037:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_IGNOLDHDR;
        break;
      case 1035:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_REFMTCARD;
        break;
      case 1036:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts &= ~OPT_REFMTCARD;
        break;
      case 1033:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_DATAOFF;
        break;
      case 1034:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts &= ~OPT_DATAOFF;
        break;
      case 1028:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_BLKOFF;
        break;
      case 1032:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts &= ~OPT_BLKOFF;
        break;
      case 1027:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_HDROFF;
        break;
      case 1031:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts &= ~OPT_HDROFF;
        break;
      case 1026:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_LINENO;
        break;
      case 1030:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts &= ~OPT_LINENO;
        break;
      case 1025:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_EMPTYLINE;
        break;
      case 1029:
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts &= ~OPT_EMPTYLINE;
        break;
      case 'n':
        if(opts & OPT_UNDEF) opts = OPT_DEFAULT;
        opts |= OPT_TEST;
        break;
      case 'v':
        if(verbose == INT_MIN)
          verbose=VERBOSE_LVL_DEFAULT+1;
        else
          ++verbose;
        break;
      case 'p':
        if(pedantic == INT_MIN)
          pedantic=1;
        else
          ++pedantic;
        break;
      case 'f':
        if(pedantic == INT_MIN)
          pedantic=-1;
        else
          --pedantic;
        break;
      case 'q':
        if(verbose == INT_MIN)
          verbose=VERBOSE_LVL_DEFAULT-1;
        else
          --verbose;
        break;
      case 'h':
        goto USAGE;
      case 'F':
        ifn = optarg;
        break;
      case 'H':
        hfn = optarg;
        break;
      default:
        rc = 126;
        goto USAGE;
    }
  }
  if(optind < (ac-1)){
    rc = 111;
    goto USAGE;
  }

  if(is_edit_mode == -1) is_edit_mode = (hfn)?1:0;
  if(optind == (ac-1)){
    if(ifn){
      fprintf(stderr, "%s: both -F (or --file) option and 'file name' argument were defined!\n", PNAME);
      return(120);
    }
    ifn = av[optind];
  }

  if(is_edit_mode){
    if(!ifn){
      fprintf(stderr, "%s: FITS file name is missed in the 'edit' mode!\n", PNAME);
      return(119);
    }
    if(!hfn) hfn = "-";
  }
  else{
    if(!ifn) ifn = "-";
  }

  if(verbose == INT_MIN) verbose = VERBOSE_LVL_DEFAULT;
  if(pedantic == INT_MIN) pedantic = CHKLVL_DEFAULT;
  if(opts & OPT_UNDEF){
    opts = OPT_DEFAULT;
    if(pedantic > CHKLVL_DEFAULT) opts |= OPT_LINENO;
  }
  else{
    if((opts&OPT_IGNOLDHDR) && pedantic > CHKLVL_DEFAULT){
      fprintf(stderr, "%s: --pedantic and --discard-old-headers! Are You joking!?\n", PNAME);
      return(98);
    }
  }

  if(!is_edit_mode){
    if(opts&OPT_IGNOLDHDR){
      fprintf(stderr, "%s: --discard-old-headers in 'dump' mode!?\n", PNAME);
      return(97);
    }
  }

  if(verbose > VERBOSE_LVL_DEBUG){
    fprintf(stderr, "%s: DEBUG: version %s\n", PNAME, FITSHDR_VERSION);
    fprintf(stderr, "%s: DEBUG: %s\n", PNAME, "$Id: fitshdr.c,v 1.69 2006/06/02 04:08:39 mik Exp $");
    fprintf(stderr, "%s: DEBUG: executed in %s mode with options:\n", PNAME, (is_edit_mode?"'edit'":"'dump'"));
    if(opts&OPT_IGNOLDHDR) fprintf(stderr, "%s: DEBUG: --discard-old-headers%s\n", PNAME, (is_edit_mode?"":" (ignored)"));
    if(opts&OPT_MODIFY_PRINCIPAL) fprintf(stderr, "%s: DEBUG: --modify-principal-keywords%s\n", PNAME, (is_edit_mode?"":" (ignored)"));
    if(opts&OPT_REFMTCARD) fprintf(stderr, "%s: DEBUG: --reformat-cards%s\n", PNAME, (is_edit_mode?"":" (ignored)"));
    if(opts&OPT_TEST) fprintf(stderr, "%s: DEBUG: --test%s\n", PNAME, (is_edit_mode?"":" (ignored)"));
    if(opts&OPT_LINENO) fprintf(stderr, "%s: DEBUG: --line-offsets%s\n", PNAME, "");
    if(opts&OPT_HDROFF) fprintf(stderr, "%s: DEBUG: --header-offsets%s\n", PNAME, (is_edit_mode?" (ignored)":""));
    if(opts&OPT_EMPTYLINE) fprintf(stderr, "%s: DEBUG: --empty-lines%s\n", PNAME, (is_edit_mode?" (ignored)":""));
    if(opts&OPT_DATAOFF) fprintf(stderr, "%s: DEBUG: --data-offsets%s\n", PNAME, (is_edit_mode?" (ignored)":""));
    if(opts&OPT_BLKOFF) fprintf(stderr, "%s: DEBUG: --block-offsets%s\n", PNAME, (is_edit_mode?" (ignored)":""));
    fprintf(stderr, "%s: DEBUG: --verbose (verbosity:%d)\n", PNAME, verbose);
    if(pedantic > CHKLVL_DEFAULT) fprintf(stderr, "%s: DEBUG: --pedantic (level:%d)\n", PNAME, pedantic);
    if(pedantic < CHKLVL_DEFAULT) fprintf(stderr, "%s: DEBUG: --force (level:%d)\n", PNAME, pedantic);
  }

  rc=((is_edit_mode)?(ed_header(ifn, hfn, opts, pedantic, verbose)):(dump_header(ifn, opts, pedantic, verbose)));

  return(rc);
USAGE:
  fprintf(stderr, "%s: Usage: %s [options] [file.fits]\n", PNAME, PNAME);
  fprintf(stderr, "To dump FITS-headers:\n");
  fprintf(stderr, "    %s file.fits\n", PNAME);
  fprintf(stderr, "    gunzip < file.fits.gz | %s\n", PNAME);
  fprintf(stderr, "To modify FITS-headers:\n");
  fprintf(stderr, "    %s -H new.hdr file.fits\n", PNAME);
  fprintf(stderr, "    %s --line-offsets --file file.fits | joe - | %s --edit file.fits\n", PNAME, PNAME);
  fprintf(stderr, "    echo \"14400:HISTORY some comments\" | %s -H - file.fits\n", PNAME);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "    -h, --help\n");
  fprintf(stderr, "        Print this message.\n");
  fprintf(stderr, "    -V, --version\n");
  fprintf(stderr, "        Print the version number.\n");
  fprintf(stderr, "    --edit\n");
  fprintf(stderr, "        'Edit' mode: in this case FITS file name has to be defined.\n");
  fprintf(stderr, "    -F \"file_name\", --fits-file=\"file_name\", --file=\"file_name\"\n");
  fprintf(stderr, "        FITS-file name, mandatory in 'edit' mode (see --edit and -H options).\n");
  fprintf(stderr, "    -H \"file_name\", --header=\"file_name\"\n");
  fprintf(stderr, "        New FITS-header file name (input). This option activates 'edit' mode.\n");
  fprintf(stderr, "    -v, --verbose\n");
  fprintf(stderr, "        Verbose mode. Multiple -v options increase the verbosity.\n");
  fprintf(stderr, "    -q, --silent\n");
  fprintf(stderr, "        Silent mode. \n");
  fprintf(stderr, "    -n, --test\n");
  fprintf(stderr, "        Test mode: don't change file ('edit' mode only).\n");
  fprintf(stderr, "    -p, --pedantic\n");
  fprintf(stderr, "        More internal checks.\n");
  fprintf(stderr, "    -f, --force\n");
  fprintf(stderr, "        Less internal checks.\n");
  fprintf(stderr, "    --[no-]empty-lines\n");
  fprintf(stderr, "        Empty lines are [not] displayed ('dump' mode only).\n");
  fprintf(stderr, "    --[no-]line-offsets\n");
  fprintf(stderr, "        Line offsets are [not] displayed/required ('dump'/'edit' mode).\n");
  fprintf(stderr, "    --[no-]header-offsets\n");
  fprintf(stderr, "        Header offsets are [not] displayed ('dump' mode only).\n");
  fprintf(stderr, "    --[no-]block-offsets\n");
  fprintf(stderr, "        Block offsets are [not] displayed ('dump' mode only).\n");
  fprintf(stderr, "    --[no-]data-offsets\n");
  fprintf(stderr, "        Data offsets are [not] displayed ('dump' mode only).\n");
  fprintf(stderr, "    --[no-]reformat-cards\n");
  fprintf(stderr, "        Re-format [or do not] input FITS header cards ('edit' mode only).\n");
  fprintf(stderr, "    --modify-principal-keywords\n");
  fprintf(stderr, "        Overwrite 'principal' FITS-keywords ('edit' mode only). !!!DANGEROUS!!!\n");
  fprintf(stderr, "    --discard-old-headers\n");
  fprintf(stderr, "        Discard/reset old FITS headers ('edit' mode only).      !!!DANGEROUS!!!\n");
  return(rc);
}

int
my_trim_tail(char *hcard, const int hlen){
  char  *cptr;
  int   i;
  
  for(i=hlen, cptr=hcard+hlen-1; i>0 && (*cptr == ' ' || *cptr == '\t'); --cptr, --i) *cptr='\0';
  return(0);  
}

int
my_convert(char *hcard, const int hlen){
  char  *cptr;
  int   i, rc;
  
  for(i=0, cptr=hcard, rc=0; i<hlen && *cptr; ++cptr, ++i)
    if(!isprint(*cptr)){
      *cptr='?';
      ++rc;
    }
  return(rc);  
}

static ssize_t
local_read(int ifileno, void *hdr_blk, size_t n){
  ssize_t nn;
  void *vptr;
  size_t  nnn;

  if(n < 0) return(-2);
  for(nnn=0, vptr=hdr_blk; nnn<n;){
    if((nn=read(ifileno, vptr, n-nnn)) == -1) return(nn);
    if(nn == 0) return(nnn);
    nnn += nn;
    vptr += nn;
  }
  return(nnn);
}

int
dump_header(const char *ifn, const int opts, const int pedantic, const int verbose){
  FILE	*f;
  int	ifileno;
  static char hdr_blk[FITS_BLK_SIZE];
  static char hcard[FITS_CARD_LEN+1];
  char	*cptr;
  static FITSHDR_ATTR hdrattr;
  static FITS_KEY fkey;
  int hdr_no, card_no, blk_no, rc, i;
  int is_end, is_first_block;
  long	l, off;

  if(ifn && strncmp(ifn, "-", sizeof("-"))){
    if((ifileno=open(ifn, O_RDONLY, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) == -1){
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: open error!\n", PNAME, ifn);
      return(124);
    }
  }
  else{
    ifileno=fileno(stdin);
  }

  f = stdout;

  fitshdr_attr_reset(&hdrattr);
  off = 0l;
  for(hdr_no=0, card_no=0, blk_no=0, is_end=0, is_first_block=1;;){
    if(is_first_block){
      if((hdrattr.hdr_off = lseek(ifileno, 0, SEEK_CUR)) != off){
        hdrattr.hdr_off = off;
      }
      hdrattr.hdr_len = 0l;
    }
    if((rc=local_read(ifileno, hdr_blk, FITS_BLK_SIZE)) != FITS_BLK_SIZE){
      if(is_first_block && !rc && off) return(0); /* just EOF */
      if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_TRUNC_HDR && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s: input FITS-file read error (truncated header?).\n", PNAME, ifn);
      return((pedantic >= CHKLVL_E_TRUNC_HDR)?64:0);
    }

    if(is_first_block){
      if(pedantic>=CHKLVL_E_MAGIC_NUM){
        if(memcmp(hdr_blk, "SIMPLE ", sizeof("SIMPLE ")-1) && \
           memcmp(hdr_blk, "XTENSION=", sizeof("XTENSION=")-1) ){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: SIMPLE or XTENSION keyword is missed: malformed header?\n", PNAME);
          if(verbose >= VERBOSE_LVL_WARNING){
            memcpy(hcard, hdr_blk, FITS_CARD_LEN);
            hcard[FITS_CARD_LEN]='\0';
            (void)my_convert(hcard, FITS_CARD_LEN);
            fprintf(stderr, "%s: card \"%ld:%-.24s...\" was found instead\n", PNAME, hdrattr.hdr_off, hcard);
          }
          return(48);
        }
      }
      if(opts & OPT_HDROFF){
        fprintf(f, "%s%ld: HEADER No.%d\n", ADDPROMPT, hdrattr.hdr_off, hdr_no);
      }
      ++hdr_no;
      is_first_block = 0;
    }
    hdrattr.hdr_len += FITS_BLK_SIZE;

    if(opts & OPT_BLKOFF){
      fprintf(f, "%s%ld: BLOCK No.%d\n", ADDPROMPT, (long)FITS_BLK_SIZE*(long)blk_no, blk_no);
    }
    
    ++blk_no;
    
    for(i=0, cptr=hdr_blk; i<FITS_BLK_SIZE/FITS_CARD_LEN; ++i, ++card_no){
      memcpy(hcard, cptr, FITS_CARD_LEN);
      hcard[FITS_CARD_LEN] = '\0';

      my_trim_tail(hcard, FITS_CARD_LEN);

      if((rc=my_convert(hcard, FITS_CARD_LEN))){
        if(pedantic >= CHKLVL_W_NONASCII){
          if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_NONASCII && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: unprintable characters in \"%ld:%-.24s...\"?\n", PNAME, hdrattr.hdr_off+card_no*FITS_CARD_LEN, hcard);
          if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: my_convert(...) == %d\n", PNAME, rc);
          if(pedantic >= CHKLVL_E_NONASCII) return(18);
        }
      }

      if(pedantic >= CHKLVL_W_CARD_FORMAT && (rc=sread_fits_key(hcard, &fkey))){
        if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_CARD_FORMAT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: header card \"%ld:%-.24s...\" has some problem(s)?\n", PNAME, hdrattr.hdr_off+card_no*FITS_CARD_LEN, hcard);
        if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: sread_fits_key(...) == %d\n", PNAME, rc);
        if(pedantic >= CHKLVL_E_CARD_FORMAT) return(17);
      }

      if(!strncmp(hcard, "END", sizeof("END"))){
        is_end = 1;
        break;
      }
      if((opts&OPT_EMPTYLINE) || *hcard){
        if(opts&OPT_LINENO){
          if((rc=fprintf(f, "%ld:", hdrattr.hdr_off+card_no*FITS_CARD_LEN)) <= 0){
            if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: fprintf: EOF?\n", PNAME, __FUNCTION__);
            return(6);
          }
        }
        if((rc=fprintf(f, "%s\n", hcard)) <= 0){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: fprintf: EOF?\n", PNAME, __FUNCTION__);
          return(5);
        }
      }
      cptr += FITS_CARD_LEN;
    }

    if((rc=fitshdr_getattr(hdr_blk, &hdrattr))){
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: header (offset=%ld) has some problems!\n", PNAME, hdrattr.hdr_off);
      if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: fitshdr_getattr(...) == %d?\n", PNAME, __FUNCTION__, rc);
      return(8);
    }

    if(is_end){
      if(opts&OPT_EMPTYLINE){
        for(i=card_no%(FITS_BLK_SIZE/FITS_CARD_LEN); i<(FITS_BLK_SIZE/FITS_CARD_LEN)-1; ++i, ++card_no){
          if((opts&OPT_LINENO) && (rc=fprintf(f, "%ld:", hdrattr.hdr_off+card_no*FITS_CARD_LEN)) <= 0){
            if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: fprintf: EOF?\n", PNAME, __FUNCTION__);
            return(18);
          }
          if((rc=fprintf(f, "\n")) <= 0){
            if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: fprintf: EOF?\n", PNAME, __FUNCTION__);
            return(18);
          }
        }
      }
      if((opts&OPT_LINENO) && (rc=fprintf(f, "%ld:", hdrattr.hdr_off+card_no*FITS_CARD_LEN)) <= 0){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: fprintf: EOF?\n", PNAME, __FUNCTION__);
        return(17);
      }
      if((rc=fprintf(f, "END\n")) <= 0){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: fprintf: EOF?\n", PNAME, __FUNCTION__);
        return(17);
      }

      if(!(hdrattr.is_completed)){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: FITS-header (offset=%ld) is not complete?!\n", PNAME, hdrattr.hdr_off);
        return(11);
      }
      if((rc=fitshdr_attr_fill(&hdrattr))){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: FITS-header (offset=%ld) is not complete?!\n", PNAME, hdrattr.hdr_off);
        if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: fitshdr_attr_fill(...) == %d?\n", PNAME, __FUNCTION__, rc);
        return(18);
      }

      off = hdrattr.hdr_off + hdrattr.hdr_len;
      if((l = ((hdrattr.datalen>0)?((hdrattr.datalen-1)/FITS_BLK_SIZE + 1):0))){
        off += l*(long)FITS_BLK_SIZE;
        if(lseek(ifileno, l*(long)FITS_BLK_SIZE, SEEK_CUR) != off){
          for(i=0; i<l; ++i){
            if((rc=local_read(ifileno, hdr_blk, FITS_BLK_SIZE)) != FITS_BLK_SIZE){
              if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: read: truncated file?\n", PNAME, __FUNCTION__);
              if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: local_read(...) == %d?\n", PNAME, __FUNCTION__, rc);
              if(rc < 0) rc = 0;
              if(opts&OPT_DATAOFF){
                fprintf(f, "%s%ld: DATA: %ld bytes (%d block(s)) != %ld bytes expected!!!\n", ADDPROMPT, hdrattr.hdr_off + hdrattr.hdr_len, (long)i*FITS_BLK_SIZE+(long)rc, i + (rc?1:0), l*FITS_BLK_SIZE);
                if((opts&OPT_BLKOFF) && (rc||i)){
                  if(i + (rc?1:0) > 1)
                    fprintf(f, "%s%ld: BLOCKs NNo.%d-%d\n", ADDPROMPT, (long)FITS_BLK_SIZE*(long)(blk_no), blk_no, blk_no + i + (rc?1:0) - 1);
                  else
                    fprintf(f, "%s%ld: BLOCK No.%d\n", ADDPROMPT, (long)FITS_BLK_SIZE*(long)(blk_no), blk_no);
                }
              }
              return(31);
            }
          }
        }
        if(opts&OPT_DATAOFF){
          fprintf(f, "%s%ld: DATA: %ld bytes (%ld block(s))\n", ADDPROMPT, hdrattr.hdr_off + hdrattr.hdr_len, l*FITS_BLK_SIZE, l);
          if(opts&OPT_BLKOFF){
            if( l > 1 )
              fprintf(f, "%s%ld: BLOCKs NNo.%d-%ld\n", ADDPROMPT, (long)FITS_BLK_SIZE*(long)(blk_no), blk_no, blk_no + l - 1);
            else
              fprintf(f, "%s%ld: BLOCK No.%d\n", ADDPROMPT, (long)FITS_BLK_SIZE*(long)(blk_no), blk_no);
          }
        }
        blk_no += l;
      }
      fitshdr_attr_reset(&hdrattr);
      card_no=0;
      is_end=0;
      is_first_block=1;
    }
  }
  return(0); /* unreachable */
}

int
find_hdr_by_off(const int ifileno, long off, FITSHDR_ATTR *hdrattr){
  static char blk[FITS_BLK_SIZE];
  long l, ll;
  int rc;

  if(!hdrattr) return(-1);

  off = off - off%FITS_BLK_SIZE;
  if(off%FITS_BLK_SIZE) return(7);

TRY_AGAIN:
  if(hdrattr->is_completed){
    if(off >= hdrattr->hdr_off){
      if(off < hdrattr->hdr_off + hdrattr->hdr_len) return(0);

      l = (hdrattr->datalen)?((hdrattr->datalen-1)/FITS_BLK_SIZE + 1):0;
      l *= FITS_BLK_SIZE;
      ll = hdrattr->hdr_off + hdrattr->hdr_len;
      if(off >= ll && off <  ll + l) return(2);
      fitshdr_attr_reset(hdrattr);
      hdrattr->hdr_off = ll + l;
    }
    else{
      fitshdr_attr_reset(hdrattr);
      hdrattr->hdr_off = 0l;
    }
  }
  else{
    fitshdr_attr_reset(hdrattr);
    hdrattr->hdr_off = 0l;
  }

  if(lseek(ifileno,  hdrattr->hdr_off, SEEK_SET) != hdrattr->hdr_off) return(-2);

  for(l=1l;;++l){
    if((rc=read(ifileno, blk, FITS_BLK_SIZE)) != FITS_BLK_SIZE) return(-3);
    if(fitshdr_getattr(blk, hdrattr)) return(-4);
    if(hdrattr->is_completed) break;
  }
  hdrattr->hdr_len = l*FITS_BLK_SIZE;
  if(fitshdr_attr_fill(hdrattr)) return(-5);
  goto TRY_AGAIN;
  
  return(0); /* unreachable */
}

static int
is_naxis_key(char *s){
  long l;
  char *cptr;

  if(!s) return(0);

  if(!strncmp(s, "NAXIS", sizeof("NAXIS")-1)){
    l = strtol(s+sizeof("NAXIS")-1, &cptr, 10);
    if(cptr != s+sizeof("NAXIS")-1 && !*cptr && l > 0 && l < 1000) return(1);
  }
  return(0);
}

static int
local_flush_block(const int ifileno, long blkoff, void *blk, int len, int opts){
  if(lseek(ifileno, blkoff, SEEK_SET) != blkoff){
    return(63);
  }
  if(!(opts&OPT_TEST) && write(ifileno, (void *)blk, len) != len){
    return(62);
  }
/*
  (void)fsync(ifileno);
 */
  return(0);
}

static int
fcard_cmp(char *c1, char *c2){
  char *cptr;
  int	i, j;

  for(i=FITS_CARD_LEN, cptr=c2+FITS_CARD_LEN-1; *cptr == ' '; --i, --cptr)
    if(i <= 1) break;

  if(*cptr == '/' ){
    --i;
    if(i) --cptr;
  }

  for(j=FITS_CARD_LEN, cptr=c1+FITS_CARD_LEN-1; *cptr == ' '; --j, --cptr)
    if(j <= 1) break;

  if(*cptr == '/' ){
    --j;
    if(j) --cptr;
  }

  if(!i) return(i?1:0);

  return(memcmp(c1, c2, ((i>j)?(i):(j))));
}

int
ed_header(const char *ifn, const char *hfn, const int opts, const int pedantic, const int verbose){
  int ifileno, mode;
  FILE *h;
  static char blk[FITS_BLK_SIZE];
  static FITSHDR_ATTR hdrattr;
  static FITS_KEY fkey, fkey_old;
  static char str[FITS_CARD_LEN+64], ostr[FITS_CARD_LEN+1], sstr[FITS_CARD_LEN+1];
  char *cptr, *cptr1;
  long off, l, blkoff, loff, boff, hlineno;
  int is_changed, with_off, rc, ii;
  static char hline[FITS_CARD_LEN+1];

  if(!strncmp(hfn, "-", sizeof("-"))){
    h=stdin;
  }
  else{
    if(!(h=fopen(hfn, "r"))){
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: 'fopen' error!\n", PNAME, hfn);
      return(123);
    }
  }

  if(ifn && strncmp(ifn, "-", sizeof("-"))){
    if(hfn){
      if(!(opts&OPT_IGNOLDHDR))
        mode = O_RDWR;
      else
        mode = (O_RDWR|O_CREAT);
    }
    else{
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%d: internal error!!\n", PNAME, __FUNCTION__, __LINE__);
      return(117);
    }
    if((ifileno=open(ifn, mode, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) == -1){
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: open error!\n", PNAME, ifn);
      return(124);
    }
  }
  else{
    if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: FITS file to edit from STDIN!?\n", PNAME);
    return(125);
  }

  fitshdr_attr_reset(&hdrattr);

  off = -1l;
  blkoff = -1l;
  is_changed = 0;
  hlineno = 0l;
  while(fgets((char *)str, sizeof(str)-1, h)){
    ++hlineno;
    str[sizeof(str)-1] = '\0';
    if((cptr=strchr(str, '\n'))){
      *cptr = '\0';
    }
    else{
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: header card \"%-.24s...\" too long!\n", PNAME, hfn, hlineno, str);
      return(96);
    }
                                           
    for(cptr=(char *)str; *cptr == ' ' || *cptr == '\t'; ++cptr) if(*cptr) break;

    if(!strncmp(cptr, COMMENT_PROMPT, sizeof(COMMENT_PROMPT)-1)){
      if(verbose > VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: DEBUG: %s:%ld: \"%-.24s...\" has been ignored.\n", PNAME, hfn, hlineno, cptr);
      continue;
    }

    l=strtol(cptr, &cptr1, 10);
    if(cptr1 != cptr && *cptr1 == ':'){
      if(*cptr == '-' || *cptr == '+'){
        if(off < 0l)
          off = 0l;
        else
          l += FITS_CARD_LEN; /* +0: offset -- the same as if offset missed */
        l += off;
        if(l < 0l){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: wrong negative offset in \"%-.16s...\"\n", PNAME, hfn, hlineno, cptr);
          return(-35);
        }
      }
      if(pedantic >= CHKLVL_E_UNSORTED && l <= off){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: offset in \"%-.16s...\" less then in previous card.\n", PNAME, hfn, hlineno, cptr);
        return(-9);
      }
      off = l;
      cptr = cptr1 + 1;
      if(opts&OPT_REFMTCARD) for(; *cptr == ' ' || *cptr == '\t'; ++cptr);
      with_off = 1;
    }
    else{
      if(opts&OPT_LINENO){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: missed offset in \"%-.16s...\"\n", PNAME, hfn, hlineno, cptr);
        return(-8);
      }

      if(!(opts&OPT_REFMTCARD)) cptr = (char *)str;

      if(off < 0l)
        off = 0l;
      else
        off += FITS_CARD_LEN;
      with_off = 0;
    }

    if(strlen(cptr) > sizeof(hline)-1){
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: header card \"%ld:%-.16s...\" too long!\n", PNAME, hfn, hlineno, off, cptr);
      return(97);
    }
    strncpy(hline, cptr, sizeof(hline)-1);
    hline[sizeof(hline)-1] = '\0';

/* input string and offset were prepared */

    if((rc=sread_fits_key(hline, &fkey))){
      if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_NEWCARD_FORMAT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s:%ld: new header card \"%ld:%-.16s...\" has some problem(s)?\n", PNAME, hfn, hlineno, off, hline);
      if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: sread_fits_key(...) == %d\n", PNAME, rc);
      if(pedantic >= CHKLVL_E_NEWCARD_FORMAT) return(10);
    }

    if(!(opts&OPT_IGNOLDHDR)){
      if((rc=find_hdr_by_off(ifileno, off, &hdrattr))){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: can't find header by offset of card!\n", PNAME, hfn, hlineno);
        if(verbose >= VERBOSE_LVL_WARNING) fprintf(stderr, "%s: current card: \"%ld:%-.16s...\"\n", PNAME, off, hline);
        if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: find_hdr_by_off(...) == %d\n", PNAME, __FUNCTION__, rc);
        return(64);
      }
      if(((boff = off)%FITS_CARD_LEN) || boff < 0l){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.16s...\" uses wrong offset?\n", PNAME, hfn, hlineno, off, hline);
        return(68);
      }

      boff -= boff%FITS_BLK_SIZE;
      if(boff < hdrattr.hdr_off || boff >= hdrattr.hdr_off + hdrattr.hdr_len){
        if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.16s...\" uses wrong offset.\n", PNAME, hfn, hlineno, off, hline);
        return(69);
      }
    }

    if((boff=off-off%FITS_BLK_SIZE) != blkoff){ /* read new block */
      if(blkoff >= 0l && is_changed){ /* flush old buffer */
        if((rc=local_flush_block(ifileno, blkoff, blk, FITS_BLK_SIZE, opts))){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: can't write header block with offset %ld!\n", PNAME, ifn, blkoff);
          if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: local_flush_block(...) == %d?\n", PNAME, __FUNCTION__, rc);
          return(63);
        }
        if(verbose > VERBOSE_LVL_DEBUG && !(opts&OPT_TEST)){
          fprintf(stderr, "%s: DEBUG: %s:%ld: block (offset=%ld) has been flushed.\n", PNAME, hfn, hlineno, blkoff);
        }
        is_changed = 0;
      }
      if(!(opts&OPT_IGNOLDHDR)){
        if((l=lseek(ifileno, boff, SEEK_SET)) != boff){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: lseek error!\n", PNAME, ifn);
          if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: lseek: %ld!=%ld?\n", PNAME, __FUNCTION__, l, boff);
          return(65);
        }

        if((rc=read(ifileno, (void *)blk, FITS_BLK_SIZE)) != FITS_BLK_SIZE){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: read error!\n", PNAME, ifn);
          if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: read: %d!=%d?\n", PNAME, __FUNCTION__, rc, FITS_BLK_SIZE);
          return(66);
        }
        if(boff == hdrattr.hdr_off+hdrattr.hdr_len-FITS_BLK_SIZE && \
           (l=FITS_BLK_SIZE-hdrattr.end_off-FITS_CARD_LEN) > 0){
          memset(blk+hdrattr.end_off+FITS_CARD_LEN, ' ', l);
        }
      }
      else{
        memset(blk, ' ', FITS_BLK_SIZE);
      }
      blkoff = boff;
    }
    
    /* check first new key */
    if(((loff=off-blkoff)%FITS_CARD_LEN) || loff >= FITS_BLK_SIZE || loff < 0l){
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: header card \"%ld:%-.16s...\" has bad offset.\n", PNAME, hfn, hlineno, off, hline);
      return(67);
    }

    if(opts&OPT_REFMTCARD){
      if(!strncmp(fkey.name, "END", sizeof("END"))){
        if(fkey.has_value || fkey.is_defined){
          if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_NEWCARD_FORMAT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s:%ld: new card \"%ld:%-.16s...\" has value?\n", PNAME, hfn, hlineno, off, hline);
          if(pedantic >= CHKLVL_E_NEWCARD_FORMAT) return(51);
	  fkey.has_value = 0;
          fkey.is_defined = 0;
        }
        if(*(fkey.comment)){
          if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_NEWCARD_FORMAT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s:%ld: new card \"%ld:%-.16s...\" has comment?\n", PNAME, hfn, hlineno, off, hline);
          if(pedantic >= CHKLVL_E_NEWCARD_FORMAT) return(52);
          *(fkey.comment) = '\0';
        }
      }

      if(fkey.has_value){
        if(!fkey.is_defined){
          if(*(fkey.comment))
            snprintf(ostr, sizeof(ostr), "%-8s= %20s / %-*s", fkey.name, " ", FITS_CARD_LEN-(10+20+3), fkey.comment);
          else
            snprintf(ostr, sizeof(ostr), "%-8s= %20s", fkey.name, " ");
        }
        else if(fkey.type == FITS_SVAL){
          sstr[sizeof(sstr)-1] = '\0';
          if(strlen(fkey.value.sval) <= 8 && strlen(fkey.value.sval))
            snprintf(sstr, sizeof(sstr)-1, "'%-8s'", fkey.value.sval);
          else
            snprintf(sstr, sizeof(sstr)-1, "'%s'", fkey.value.sval);
          sstr[sizeof(sstr)-1] = '\0';

          (void)my_trim_tail(fkey.comment, sizeof(fkey.comment)-1);
          

          if(strlen(sstr) > FITS_CARD_LEN-10){
            if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: header card \"%ld:%-.16s...\" too long!\n", PNAME, hfn, hlineno, off, hline);
            return(90);
          }

          if((ii=strlen(sstr)) < 20) ii = 20;
          if(ii+strlen(fkey.comment)+3 > FITS_CARD_LEN-10){
            if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_TRUNC_COMMENT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.16s...\": too long comment?!\n", PNAME, hfn, hlineno, off, hline);
            if(pedantic >= CHKLVL_E_TRUNC_COMMENT) return(91);
          }

          if(strlen(sstr) > 20){
            if(*(fkey.comment))
              snprintf(ostr, sizeof(ostr), "%-8s= %s / %s", fkey.name, sstr, fkey.comment);
            else
              snprintf(ostr, sizeof(ostr), "%-8s= %s", fkey.name, sstr);
          }
          else{
            if(*(fkey.comment))
              snprintf(ostr, sizeof(ostr), "%-8s= %-20s / %-*s", fkey.name, sstr, FITS_CARD_LEN-(10+20+3), fkey.comment);
            else
              snprintf(ostr, sizeof(ostr), "%-8s= %-20s", fkey.name, sstr);
          }
        }
        else if(fkey.type == FITS_BVAL){
          if(*(fkey.comment))
            snprintf(ostr, sizeof(ostr), "%-8s= %20s / %-*s", fkey.name, fkey.value.bval?"T":"F", FITS_CARD_LEN-(10+20+3), fkey.comment);
          else
            snprintf(ostr, sizeof(ostr), "%-8s= %20s", fkey.name, fkey.value.bval?"T":"F");
        }
        else if(fkey.type == FITS_IVAL){
          if(*(fkey.comment))
            snprintf(ostr, sizeof(ostr), "%-8s= %20ld / %-*s", fkey.name, fkey.value.ival, FITS_CARD_LEN-(10+20+3), fkey.comment);
          else
            snprintf(ostr, sizeof(ostr), "%-8s= %20ld", fkey.name, fkey.value.ival);
        }
        else if(fkey.type == FITS_FVAL){
          snprintf(sstr, 19, "%.17G", fkey.value.fval);
          sstr[19-1] = '\0';
          sstr[sizeof(sstr)-1] = '\0';

          if(!strchr(sstr, '.')) strcat(sstr, ".0");
          if(*(fkey.comment))
            snprintf(ostr, sizeof(ostr), "%-8s= %20s / %-*s", fkey.name, sstr, FITS_CARD_LEN-(10+20+3), fkey.comment);
          else
            snprintf(ostr, sizeof(ostr), "%-8s= %20s", fkey.name, sstr);
        }
        else{
          if(strlen(fkey.comment) > FITS_CARD_LEN - 9){
            if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_TRUNC_COMMENT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.16s...\": too long comment?!\n", PNAME, hfn, hlineno, off, hline);
            if(pedantic >= CHKLVL_E_TRUNC_COMMENT) return(53);
          }
          snprintf(ostr, sizeof(ostr), "%-8s %s", fkey.name, fkey.comment);
        }
        ostr[sizeof(ostr)-1] = '\0';
        ostr[FITS_CARD_LEN] = '\0';
      }
      else{
        if(strlen(fkey.comment) > FITS_CARD_LEN - 9){
          if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_TRUNC_COMMENT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.16s...\": too long comment?!\n", PNAME, hfn, hlineno, off, hline);
          if(pedantic >= CHKLVL_E_TRUNC_COMMENT) return(54);
        }
        snprintf(ostr, sizeof(ostr), "%-8s %s", fkey.name, fkey.comment);
        ostr[sizeof(ostr)-1] = '\0';
      }
    }
    else{
      if(strlen(hline) > FITS_CARD_LEN){
        if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_TRUNC_COMMENT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.16s...\": too long comment?!\n", PNAME, hfn, hlineno, off, hline);
        if(pedantic >= CHKLVL_E_TRUNC_COMMENT) return(55); /* does not work! */
      }
      strncpy(ostr, hline, sizeof(ostr)-1);
      ostr[sizeof(ostr)-1] = '\0';
    }

    for(cptr1=ostr+strlen(ostr); cptr1-ostr < FITS_CARD_LEN;) *cptr1++ = ' ';
    ostr[FITS_CARD_LEN] = '\0';

/* new card is ready now */
/* better to finish with "END" cards (new and old) 1st */
/* and only after check if cards (new and old) are identical */

    if(!(opts&OPT_IGNOLDHDR)){
      if((rc=sread_fits_key(blk+loff, &fkey_old))){
        if(verbose >= VERBOSE_LVL_WARNING || (pedantic >= CHKLVL_E_OLDCARD_FORMAT && verbose >= VERBOSE_LVL_ERROR)) fprintf(stderr, "%s: %s: old header card \"%ld:%-8s...\" has some problem(s).\n", PNAME, ifn, off, fkey_old.name);
        if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: sread_fits_key(...) == %d\n", PNAME, rc);
        if(pedantic >= CHKLVL_E_OLDCARD_FORMAT) return(9);
      }
    }

    if(!strncmp(fkey.name, "END", sizeof("END"))){
      if(!(opts&OPT_IGNOLDHDR)){
        if(strncmp(fkey_old.name, "END", sizeof("END")) && with_off){
          if(blkoff+FITS_BLK_SIZE != hdrattr.hdr_off+ hdrattr.hdr_len){
            if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: header card \"%ld:END\" has wrong offset!\n", PNAME, hfn, hlineno, off);
            return(72);
          }

          if((l=hdrattr.end_off%FITS_BLK_SIZE) < loff) (void)memset(blk+l, ' ', FITS_CARD_LEN);
          (void)memcpy(blk+loff, ostr, FITS_CARD_LEN);
          if((l=FITS_BLK_SIZE-loff-FITS_CARD_LEN)) (void)memset(blk+loff+FITS_CARD_LEN, ' ', l);
          hdrattr.end_off = loff;
          is_changed=1;

          if(verbose > VERBOSE_LVL_DEBUG){
            fprintf(stderr, "%s: DEBUG: %s:%ld: \"%ld:%-.24s...\" has been buffered.\n", PNAME, hfn, hlineno, off, hline);
          }

          off = (hdrattr.datalen)?((hdrattr.datalen-1)/FITS_BLK_SIZE + 1):0;
          off *= FITS_BLK_SIZE;
          off += hdrattr.hdr_off + hdrattr.hdr_len;
          off -= FITS_CARD_LEN;
        }
        else{
          off = (hdrattr.datalen)?((hdrattr.datalen-1)/FITS_BLK_SIZE + 1):0;
          off *= FITS_BLK_SIZE;
          off += hdrattr.hdr_off + hdrattr.hdr_len;
          if(verbose > VERBOSE_LVL_DEBUG){
            fprintf(stderr, "%s: DEBUG: %s:%ld: \"%.9s...\": expected offset was changed to %ld, card was skipped.\n", PNAME, hfn, hlineno, hline, off);
          }
          off -= FITS_CARD_LEN;
        }
      }
      else{
        (void)memcpy(blk+loff, ostr, FITS_CARD_LEN);
        if(verbose > VERBOSE_LVL_DEBUG){
          fprintf(stderr, "%s: DEBUG: %s:%ld: \"%ld:%-.24s...\" has been buffered.\n", PNAME, hfn, hlineno, off, hline);
        }

        if((l=FITS_BLK_SIZE - loff - FITS_CARD_LEN)) (void)memset(blk+loff+FITS_CARD_LEN, ' ', l);
        if((rc=local_flush_block(ifileno, blkoff, blk, FITS_BLK_SIZE, opts))){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: can't write header block with offset %ld!\n", PNAME, ifn, blkoff);
          if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: local_flush_block(...) == %d?\n", PNAME, __FUNCTION__, rc);
          return(84);
        }
        if(verbose > VERBOSE_LVL_DEBUG && !(opts&OPT_TEST)){
          fprintf(stderr, "%s: DEBUG: %s:%ld: block (offset=%ld) has been flushed.\n", PNAME, hfn, hlineno, blkoff);
        }

        if((rc=find_hdr_by_off(ifileno, off, &hdrattr))){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: not enough information to find next header.\n", PNAME, hfn, hlineno);
          if(verbose >= VERBOSE_LVL_WARNING) fprintf(stderr, "%s: %s: current line: \"%ld:%-.16s\"\n", PNAME, __FUNCTION__, off, hline);
          if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: find_hdr_by_off(...) == %d\n", PNAME, __FUNCTION__, rc);
          return(85);
        }
        off = (hdrattr.datalen)?((hdrattr.datalen-1)/FITS_BLK_SIZE + 1):0;
        off *= FITS_BLK_SIZE;
        off += hdrattr.hdr_off + hdrattr.hdr_len;
	off -= FITS_CARD_LEN;
        is_changed=0;
      }
      continue;
    }

    if(!(opts&OPT_IGNOLDHDR)){
      if(!strncmp(fkey_old.name, "END", sizeof("END")) || \
         off >= hdrattr.hdr_off + hdrattr.hdr_len - FITS_BLK_SIZE + hdrattr.end_off){
        if(FITS_BLK_SIZE-loff < FITS_CARD_LEN*2){
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: not enough space to add card \"%ld:%-.16s...\"\n", PNAME, hfn, hlineno, off, hline);
          return(73);
        }

        (void)memcpy(blk+loff+FITS_CARD_LEN, blk+hdrattr.end_off, FITS_CARD_LEN);
        (void)memset(blk+hdrattr.end_off, ' ', FITS_CARD_LEN);
        (void)memcpy(blk+loff, ostr, FITS_CARD_LEN);
        hdrattr.end_off = loff+FITS_CARD_LEN;
        if(verbose > VERBOSE_LVL_DEBUG){
          fprintf(stderr, "%s: DEBUG: %s:%ld: \"%ld:%-.24s...\" has been buffered, \"END\" card was moved.\n", PNAME, hfn, hlineno, off, hline);
        }
        is_changed=1;
        continue;
      }
    }

/* check, if identical */
    if(!(opts&OPT_IGNOLDHDR)){
      if(!fcard_cmp(blk+loff, ostr)){
        if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.15s...\" is identical to the original one, ignored!\n", PNAME, hfn, hlineno, off, hline);
        continue;
      }
      if(fkey_old.is_defined && fkey.is_defined){
        if(!strcmp(fkey_old.name, fkey.name) && \
           fkey_old.type == fkey.type && \
           !strcmp(fkey_old.comment, fkey.comment) &&
           ( (fkey.type == FITS_SVAL && !strcmp(fkey.value.sval, fkey_old.value.sval)) || \
             (fkey.type == FITS_IVAL && fkey.value.ival == fkey_old.value.ival) || \
             (fkey.type == FITS_FVAL && fabs(fkey.value.fval - fkey_old.value.fval)<1e-15) || \
             (fkey.type == FITS_BVAL && fkey.value.bval == fkey_old.value.bval) \
           ) ){
          if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.15s...\" is identical (quasi) to the original one, ignored!\n", PNAME, hfn, hlineno, off, hline);
          continue;
        }
      }
      if(!fkey_old.is_defined && !fkey.is_defined && \
          fkey.type == FITS_NOVAL && fkey_old.type == FITS_NOVAL){
        if(!strcmp(fkey_old.comment, fkey.comment) && !strcmp(fkey_old.name, fkey.name)){
          if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s:%ld: card \"%ld:%-.15s...\" is very similar to the original one, ignored!\n", PNAME, hfn, hlineno, off, hline);
          continue;
        }
      }
    }
    
    if(!(opts&OPT_MODIFY_PRINCIPAL)){
      if(!(opts&OPT_IGNOLDHDR)){
        if(fkey_old.is_defined){ /* probably, wrong check */
          if(!strncmp(fkey_old.name, "NAXIS", sizeof("NAXIS"))   || \
             !strncmp(fkey_old.name, "PCOUNT", sizeof("PCOUNT")) || \
             !strncmp(fkey_old.name, "GCOUNT", sizeof("GCOUNT")) || \
             !strncmp(fkey_old.name, "BITPIX", sizeof("BITPIX")) || \
             !strncmp(fkey_old.name, "GROUPS", sizeof("GROUPS")) || \
             !strncmp(fkey_old.name, "SIMPLE", sizeof("SIMPLE")) || \
             !strncmp(fkey_old.name, "XTENSION", sizeof("XTENSION")) || \
             is_naxis_key((fkey_old.name)) ){ /* skip this keys */
            if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: header card \"%ld:%-.16s...\" has conflict with old one!\n", PNAME, hfn, hlineno, off, hline);
            if(verbose >= VERBOSE_LVL_WARNING) fprintf(stderr, "%s: %s:%ld: old header card: \"%ld:%-8s...\" is 'principal'!\n", PNAME, hfn, hlineno, off, fkey_old.name);
            return(101);
          }
        }
      }

      if(!(opts&OPT_IGNOLDHDR) && pedantic >= CHKLVL_CHK_NEWCARD){
        if(!strncmp(fkey.name, "NAXIS", sizeof("NAXIS"))   || \
           !strncmp(fkey.name, "PCOUNT", sizeof("PCOUNT")) || \
           !strncmp(fkey.name, "GCOUNT", sizeof("GCOUNT")) || \
           !strncmp(fkey.name, "BITPIX", sizeof("BITPIX")) || \
           !strncmp(fkey.name, "GROUPS", sizeof("GROUPS")) || \
           !strncmp(fkey.name, "SIMPLE", sizeof("SIMPLE")) || \
           !strncmp(fkey.name, "XTENSION", sizeof("XTENSION")) || \
           is_naxis_key((fkey.name)) ){ /* skip this keys also */
          if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s:%ld: header card \"%ld:%-.16s...\" might change FITS-file structure!\n", PNAME, hfn, hlineno, off, hline);
          return(102);
        }
      }
    }

    (void)memcpy(blk+loff, ostr, FITS_CARD_LEN);
    if(verbose > VERBOSE_LVL_DEBUG){
      fprintf(stderr, "%s: DEBUG: %s:%ld: \"%ld:%-.24s...\" has been buffered.\n", PNAME, hfn, hlineno, off, hline);
    }
    is_changed=1;
  }

  if(blkoff >= 0l && is_changed){ /* flush old buffer */
    if((rc=local_flush_block(ifileno, blkoff, blk, FITS_BLK_SIZE, opts))){
      if(verbose >= VERBOSE_LVL_ERROR) fprintf(stderr, "%s: %s: can't write header block with offset %ld!\n", PNAME, ifn, blkoff);
      if(verbose >= VERBOSE_LVL_DEBUG) fprintf(stderr, "%s: %s: local_flush_block(...) == %d?\n", PNAME, __FUNCTION__, rc);
      return(83);
    }
    if(verbose > VERBOSE_LVL_DEBUG && !(opts&OPT_TEST)){
      fprintf(stderr, "%s: DEBUG: %s:%ld: block (offset=%ld) has been flushed.\n", PNAME, hfn, hlineno, blkoff);
    }
    is_changed = 0;
  }

  (void)close(ifileno);
  return(0);
}

/* ChangeLog....
 * $Log: fitshdr.c,v $
 * Revision 1.69  2006/06/02 04:08:39  mik
 * more "clever" checking for identical header lines...
 *
 * Revision 1.68  2005/08/25 10:14:56  mik
 * several function were added in ../common/fits.c (is not used in fitshdr)...
 *
 * Revision 1.67  2005/01/31 12:36:53  mik
 * ver.1.0.4 rel.2: cosmetic changes...
 *
 * Revision 1.66  2005/01/31 12:18:16  mik
 * ...diagnostic message was changed...
 *
 * Revision 1.65  2005/01/31 12:11:58  mik
 * ...new option: --modify-principal-keywords...
 *
 * Revision 1.64  2005/01/21 18:16:18  mik
 * ver.1.0.4 rel.1: comparison of comments for header lines was changed...
 *
 * Revision 1.63  2004/12/06 17:50:02  mik
 * ...error, previous correction related...
 *
 * Revision 1.62  2004/12/06 17:40:12  mik
 * ver.1.0.3 rel.2: "END" card forced to flush last block even if no changes. fixed
 *
 * Revision 1.61  2004/11/25 10:04:58  mik
 * ...priorities of checks were changed... (--pedantic)...
 *
 * Revision 1.60  2004/11/24 17:56:05  mik
 * ver.1.0.3: undefined values (stupid standard specification!) are supported....
 *
 * Revision 1.59  2004/11/24 15:36:06  mik
 * ...empty string value formatting...
 *
 * Revision 1.58  2004/11/15 15:37:41  mik
 * ...problem with empty comment field for BITPIX,NAXIS,...
 *
 * Revision 1.57  2004/10/27 14:52:29  mik
 * ...lint, again...
 *
 * Revision 1.56  2004/10/27 14:36:35  mik
 * ver.1.0.2: ...lint...
 *
 * Revision 1.55  2004/10/27 13:58:57  mik
 * ...stipid error in usage message: sometimes can stop execitons...
 *
 * Revision 1.54  2004/10/27 13:47:49  mik
 * ...compilation under Sun OS...
 *
 * Revision 1.53  2004/10/27 09:43:02  mik
 * ...error in usage message, again...
 *
 * Revision 1.52  2004/10/26 10:06:37  mik
 * ...small correction in usage message...
 *
 * Revision 1.51  2004/10/26 09:13:06  mik
 * ver.1.0.1: ...bit different way to handle comments...
 *
 * Revision 1.50  2004/10/25 10:43:54  mik
 * ver.1.0b10: more pedantic checks... critical FITS-keys related...
 *
 * Revision 1.49  2004/10/20 18:03:30  mik
 * ver.1.0b9: +0: offset is the same as if offset is missed...
 *
 * Revision 1.48  2004/09/27 12:16:04  mik
 * ...problem with long comment line... (keys without value only)
 *
 * Revision 1.47  2004/09/27 10:35:03  mik
 * ...report as error empty file in 'dump'-mode...
 *
 * Revision 1.46  2004/09/25 19:15:20  mik
 * ...small correction in debug message...
 *
 * Revision 1.45  2004/09/25 19:01:57  mik
 * ...serious problem with "END" card (without offset) was(?) fixed...
 *
 * Revision 1.44  2004/09/25 17:50:38  mik
 * ...more debug info...
 *
 * Revision 1.43  2004/09/25 16:42:00  mik
 * ver.1.0b8: DEBUG messages... bug related to --discard-old-headers was fixed...
 *
 * Revision 1.42  2004/09/23 10:07:45  mik
 * ...internal chesk priority was changed...
 *
 * Revision 1.41  2004/09/22 18:08:06  mik
 * ...cosmetics...
 *
 * Revision 1.40  2004/09/22 17:00:57  mik
 * ver.1.0b7: formating of comment-like input cards...
 *
 * Revision 1.39  2004/09/22 14:18:15  mik
 * ...cosmetics...
 *
 * Revision 1.38  2004/09/22 14:12:53  mik
 * ...one warning message was changed...
 *
 * Revision 1.37  2004/09/22 13:47:13  mik
 * ver.1.0b6: order of internal checks was changed...
 *
 * Revision 1.36  2004/09/22 11:41:20  mik
 * ...diagnostic messages were changed...
 *
 * Revision 1.35  2004/09/21 09:39:44  mik
 * ver.1.0b5... .: ofsets were removed, +nnn:, -nnn: were added instead...
 *
 * Revision 1.34  2004/09/20 17:07:55  mik
 * ...error message was corrected...
 *
 * Revision 1.33  2004/09/20 15:48:28  mik
 * ...more check for options...
 *
 * Revision 1.32  2004/09/20 11:45:26  mik
 * ...default options...
 *
 * Revision 1.31  2004/09/17 16:06:22  mik
 * ...cosmetics...
 *
 * Revision 1.30  2004/09/17 16:02:00  mik
 * help message was changed...
 *
 * Revision 1.29  2004/09/17 14:25:31  mik
 * ...more clear error message...
 *
 * Revision 1.28  2004/09/17 14:13:16  mik
 * ...a lot of changes, mainly -- cmdline interface & diagnostic messages...
 *
 * Revision 1.27  2004/09/16 15:39:12  mik
 * ...internal checks & priorities, again...
 *
 * Revision 1.26  2004/09/16 15:27:25  mik
 * ...priorities, again...
 *
 * Revision 1.25  2004/09/16 15:25:23  mik
 * ...priorities for internal checks were changed...
 *
 * Revision 1.24  2004/09/16 15:20:09  mik
 * ...cosmetics...
 *
 * Revision 1.23  2004/09/16 14:50:44  mik
 * ...problem with nonascii characters checking was fixed...
 *
 * Revision 1.22  2004/09/16 13:33:24  mik
 * ...more pedantic work with strings and comments...
 *
 * Revision 1.21  2004/09/16 11:06:03  mik
 * ...wrong block counter in case of data blocks...
 *
 * Revision 1.20  2004/09/16 10:48:53  mik
 * ...block information output was changed...
 *
 * Revision 1.19  2004/09/16 10:17:04  mik
 * ...wrong bytes counter in case of truncated data...
 *
 * Revision 1.18  2004/09/15 13:18:00  mik
 * ...version 1.0b1: bug fix... plus few new features...
 *
 * Revision 1.17  2004/09/14 14:14:27  mik
 * ...cmdline interface was changed... a bit...
 *
 * Revision 1.16  2004/09/14 13:52:10  mik
 * help message was changed...
 *
 * Revision 1.15  2004/09/14 11:24:39  mik
 * ...version information was changed again...
 *
 * Revision 1.14  2004/09/14 11:16:36  mik
 * ...version info was changed...
 *
 * Revision 1.13  2004/09/14 11:10:27  mik
 * ...default options were changed...
 *
 * Revision 1.12  2004/08/27 14:42:23  mik
 * ...bug affected --no-li editing was fixed.
 *
 * Revision 1.11  2004/08/27 13:43:03  mik
 * ...deep modification...
 *
 * Revision 1.10  2004/08/26 13:44:07  mik
 * ...more relaxed mode for editing was added...
 *
 * Revision 1.9  2004/08/25 16:24:55  mik
 * several new checks were added...
 *
 * Revision 1.8  2004/08/25 16:02:25  mik
 * ...more internal checks...
 *
 * Revision 1.7  2004/08/25 14:52:32  mik
 * more examples in help...
 *
 * Revision 1.6  2004/08/25 14:14:45  mik
 * small modification in 'local_read'...
 *
 * Revision 1.5  2004/08/25 13:59:48  mik
 * MEF & stdio/pipe related bugs were fixed...
 *
 * Revision 1.4  2004/08/24 21:11:34  mik
 * new command line options were added...
 *
 * Revision 1.3  2004/08/24 20:53:56  mik
 * 1st beta version...
 *
 * Revision 1.2  2004/08/20 12:20:59  mik
 * 1st commited version...
 */
